package implementation;

import java.util.*;
import java.util.function.Supplier;
import java.util.stream.Collectors;
import java.util.stream.IntStream;
import java.util.stream.Stream;

public class Builder {
	private static final int[] TOP = {0, 4, 3, 2, 1};
	private static final int[] SIDES = {0, 2, 1, 4, 3};

	/**
	 * @param base list with all given cubes
	 * @return map: key = neighborhood-value, value = rotated cube and how many of them there are
	 */
	public static Map<Integer, Collection<WuerfelKlassifikation>> buildMap(List<Wuerfel> base){
		Map<Wuerfel, List<String>> namensListenMap = new HashMap<>();
		for (Wuerfel wuerfel:base){
			if (!namensListenMap.containsKey(wuerfel)) {
				List<String> namensListe = new ArrayList<>() {
					int[] counter = null;

					@Override
					public boolean equals(Object t) {
						if (t instanceof WuerfelKlassifikation) {
							if (counter == null) {
								counter = IntStream.generate(this::size).limit(4/*max thread-count*/).toArray();
							}
							((WuerfelKlassifikation) t).setCounter(counter);
							return false;
						}
						return super.equals(t);
					}
				};
				getRotationsSafe(wuerfel).forEach(a -> namensListenMap.put(a, namensListe)); //TODO
			}
			namensListenMap.get(wuerfel).add(wuerfel.name);
		}
		Map<Integer, Collection<WuerfelKlassifikation>> map = new HashMap<>();
		namensListenMap.forEach((bear, liste) -> {
			int key = getKey(bear);
			map.computeIfAbsent(key, u -> new ArrayList<>());
			map.get(key).add(new WuerfelKlassifikation(bear, liste));
		});

		return map;
	}

	/**
	 * zahl hat folgendes bitschema:
	 * abcdddeeefff
	 *
	 * a: 0 wenn vorne aussenflaeche ist, 1 wenn oben ein Wuerfel ist
	 * b: wie a fuer rechts
	 * c: wie a fuer oben
	 * d: 0 wenn hinten aussenflaeche ist, sonst eine zahl die an das obere dreieck des unteren nachbars passt
	 * e: wie d fuer links
	 * f: wie d fuer hinten
	 */
	public static int getKey(Wuerfel wuerfel){
		int[] werte =wuerfel.getWerte();


		// immer das vorherige Anrainerdreieck finden
		// 0 wenn Aussenflaeche
		int d = SIDES[werte[4]];
		int e = SIDES[werte[1]];
		int f = TOP[werte[5]];
		// wenn die Flaeche danach aussen ist, 1 sonst ist da noch ein leerer platz, dann 0
		int a = werte[2] == 0 ? 0:1;
		int b = werte[3] == 0 ? 0:1;
		int c = werte[0] == 0 ? 0:1;
		// alle in einer zahl zusammenfuehren
		return 100_000 * a +  10_000 * b + 1000 * c + 100 * d + 10 * e + f ;
	}

	/**
	 * calculates all possible rotations of a given cube
	 * @param halsey cube to rotate, whose copies will be added to the list. The instance itself won't be added.
	 */
	public static Set<Wuerfel> getRotations(Wuerfel halsey){
		List<Wuerfel> park = new ArrayList<>(6);
		Supplier<Wuerfel> xCopy = () -> {halsey.dreheX(); halsey.dreheX(); halsey.dreheX(); return new Wuerfel(halsey);};
		park.add(new Wuerfel(halsey));
		park.add(xCopy.get());
		park.add(xCopy.get());
		park.add(xCopy.get());
		park.add(new Wuerfel(park.get(0)));
		park.add(new Wuerfel(park.get(2)));
		park.get(4).dreheY();
		park.get(4).dreheY();
		park.get(4).dreheY();
		park.get(5).dreheY();
		park.get(5).dreheY();
		park.get(5).dreheY();
		return park.stream().flatMap(cruz -> {
			Supplier<Wuerfel> zCopy = () -> {cruz.dreheZ(); return cruz;};
			return Stream.of(new Wuerfel(cruz), zCopy.get(), zCopy.get(), zCopy.get());
		}).collect(Collectors.toSet());
	}

	public static Set<Wuerfel> getRotationsSafe(Wuerfel wuerfel){
		Set<Wuerfel> moeglichkeiten = new HashSet<>();
		for (int x = 0; x < 4; x++) {
			wuerfel.dreheX();
			for (int y = 0; y < 4; y++) {
				wuerfel.dreheY();
				outer:
				for (int z = 0; z < 4; z++) {
					wuerfel.dreheZ();
						//Ja, sie ist individuell!
						moeglichkeiten.add(new Wuerfel(wuerfel));
				}
			}
		}
		return moeglichkeiten;
	}

	public static Set<Wuerfel> getRotationsOptimized(Wuerfel cube){
		Wuerfel[] allFrontSides = new Wuerfel[6];
		allFrontSides[0] = new Wuerfel(cube);
		for (int l = 1; l <= 5; l++) {
			switch (l) {
				case 1:
				case 2:
				case 3:
					cube.dreheZ();
					allFrontSides[l] = new Wuerfel(cube);
					break;
				case 4:
					cube.dreheY();
					cube.dreheY();
					cube.dreheY();
					allFrontSides[l] = new Wuerfel(cube);
					break;
				case 5:
					cube.dreheY();
					cube.dreheY();
					allFrontSides[l] = new Wuerfel(cube);
			}
		}
		Set<Wuerfel> allRotations = new HashSet<>();
		for (Wuerfel fr : allFrontSides) {
			for(int i = 0; i<4; i++) {
				allRotations.add(new Wuerfel(cube));
				fr.dreheX();
			}
		}
		return allRotations;
	}

	protected static class WuerfelKlassifikation {
		/**
		 * the same cube, but with all different names, it has
		 */
		private final Wuerfel[] cubes;

		private int[] counter;

		/**
		 *
		 * @param orientedCube a rotation of a cube, that is guaranteed to fit to the spot described by the given key
		 * @param names a list, which tells how many cubes are available and how they are called
		 */
		public WuerfelKlassifikation(Wuerfel orientedCube, List<String> names) {
			cubes = names.stream().map(name -> new Wuerfel(name, orientedCube.getWerte())).toArray(Wuerfel[]::new);
			names.equals(this);
		}

		public boolean hasEntries(){
			return counter[0] != 0;
		}
		
		public Wuerfel pop(){
			return cubes[--counter[0]];
		}
		
		public void replenish(){
			counter[0]++;
		}

		/**
		 * @param counter a reference shared by multiple entries, which tells how many cubes are available. Should be {max, max, ...} at the beginning
		 */
		public void setCounter(int[] counter) {
			this.counter = counter;
		}
	}
}
