package implementation;

import java.util.*;
import java.util.function.Function;

public class PuzzleLoeser {
	//Attribute
	private HolzPuzzle holzPuzzle;
	private HashMap<Integer, Integer> fitsMap;

	public PuzzleLoeser() {
		fitsMap = new HashMap<>();
		fitsMap.put(0, 0);
		fitsMap.put(1, 4);
		fitsMap.put(2, 3);
		fitsMap.put(3, 2);
		fitsMap.put(4, 1);
		fitsMap.put(10, 0);
		fitsMap.put(11, 2);
		fitsMap.put(12, 1);
		fitsMap.put(13, 4);
		fitsMap.put(14, 3);
		fitsMap.put(20, 0);
		fitsMap.put(21, 2);
		fitsMap.put(22, 1);
		fitsMap.put(23, 4);
		fitsMap.put(24, 3);
		fitsMap.put(30, 0);
		fitsMap.put(31, 2);
		fitsMap.put(32, 1);
		fitsMap.put(33, 4);
		fitsMap.put(34, 3);
		fitsMap.put(40, 0);
		fitsMap.put(41, 2);
		fitsMap.put(42, 1);
		fitsMap.put(43, 4);
		fitsMap.put(44, 3);
		fitsMap.put(50, 0);
		fitsMap.put(51, 4);
		fitsMap.put(52, 3);
		fitsMap.put(53, 2);
		fitsMap.put(54, 1);
	}

	/**
	 * Verarbeitet das Holzpuzzle und versucht es durch geschicktes Ausschliessen zu loesen
	 *
	 * @param holzPuzzle ungeloestes Holzpuzzle
	 * @return geloestes Holzpuzzle oder null (bei keiner Loesung)
	 */
	public HolzPuzzle process(HolzPuzzle holzPuzzle) {
		this.holzPuzzle = holzPuzzle;
		List<Wuerfel>[] klassen = klassifiziere();
		if (!Loesbarkeitspruefung.isSolvableRegardingClasses(this.holzPuzzle, klassen)) {
			return null;
		}
		return belegePuzzle(holzPuzzle.getQuader(), new int[]{0, 0, 0}, klassen);
	}

	/**
	 * Klassifiziere die Wuerfel aus der Liste anhand ihrer Aussenseiten
	 *
	 * @return Array aus Wuerfelklassen
	 */
	private List<Wuerfel>[] klassifiziere() {
		List<Wuerfel>[] klassen = new List[7];
		for (int i = 0; i < 7; i++) {
			klassen[i] = new ArrayList<>();
		}
		for (Wuerfel wuerfel : this.holzPuzzle.getTeile()) {
			klassen[wuerfel.getAussenflaechen()].add(wuerfel);
		}
		return klassen;
	}

	/**
	 * Rekursive Methode. Belegt den quader mit den Wuerfeln in wuerfel, startend von position
	 *
	 * @param quader   Form des Holzpuzzles
	 * @param position x,y,z-Indizes des einzufuegenden Wuerfels
	 * @param wuerfel  Klassen von wuerfeln, die noch eingefuegt werden muessen
	 * @return korrekt geloester Wuerfel oder null
	 */
	private HolzPuzzle belegePuzzle(Wuerfel[][][] quader, int[] position, List<Wuerfel>[] wuerfel) {
		//Ermittle Schema
		int[] maske = getMaske(quader, position);
		//Ermittle benoetigte Klasse
		int aussenCounter = 0;
		for (int maskenWert : maske) {
			if (maskenWert == 0) {
				aussenCounter++;
			}
		}
		//Gehe alle Wuerfel aus der Klasse durch
		List<Wuerfel> elemente = wuerfel[aussenCounter];
		for (Wuerfel element : elemente) {
			//Betrachtetes Element aus den Klassen entfernen
			List<Wuerfel>[] klassenNeu = Arrays.stream(wuerfel).map((Function<List<Wuerfel>, ArrayList>) ArrayList::new).toArray(List[]::new);
			klassenNeu[aussenCounter].remove(element);
			//Wuerfel drehen und schauen ob er passt
			List<Wuerfel> moeglichkeiten = fit(element, maske);
			//Alle Moeglichkeiten durchgehen
			for (Wuerfel m : moeglichkeiten) {
				Wuerfel[][][] quaderNeu = Arrays.stream(quader).map(Arrays::stream).map(x -> x.map(Arrays::stream).map(y -> y.toArray(Wuerfel[]::new))
					.toArray(Wuerfel[][]::new)).toArray(Wuerfel[][][]::new);
				quaderNeu[position[0]][position[1]][position[2]] = m;
				//Check auf volles quader
				int elementCounter = 0;
				for (List<Wuerfel> klasse : klassenNeu) {
					elementCounter += klasse.size();
				}
				if (elementCounter == 0) {
					//Wir haben eine Loesung!
					return new HolzPuzzle(this.holzPuzzle.getTeile(), quaderNeu, this.holzPuzzle.getKommentare());
				}
				HolzPuzzle ergebnis = belegePuzzle(quaderNeu, findPosition(quaderNeu, position), klassenNeu);
				if (ergebnis != null) {
					return ergebnis;
				}
			}
		}
		return null;
	}

	/**
	 * Bestimmt alle Wuerfeldrehungen eines Wuerfels, die in den Quader passen
	 *
	 * @param wuerfel betrachteter Wuerfel
	 * @param maske   Schema, in das der Wuerfel passen muss
	 * @return Liste aller Moeglichkeiten, wie der Wuerfel passend gedreht werden kann
	 */
	private List<Wuerfel> fit(Wuerfel wuerfel, int[] maske) {
		List<Wuerfel> moeglichkeiten = new ArrayList<>();
		for (int x = 0; x < 4; x++) {
			wuerfel.dreheX();
			for (int y = 0; y < 4; y++) {
				wuerfel.dreheY();
				outer:
				for (int z = 0; z < 4; z++) {
					wuerfel.dreheZ();
					if (fits(wuerfel, maske)) {
						//Wir haben eine Moeglichkeit gefunden, aber ist sie schon in der Loesungsmenge?
						for (Wuerfel w : moeglichkeiten) {
							if (w.toString().equals(wuerfel.toString())) {
								continue outer;
							}
						}
						//Ja, sie ist individuell!
						moeglichkeiten.add(new Wuerfel(wuerfel));
					}
				}
			}
		}
		return moeglichkeiten;
	}

	/**
	 * Bestimmt ob der Wuerfel in die Vorgabe passt
	 *
	 * @param wuerfel betrachteter Wuerfel
	 * @param maske   schematische Vorgabe
	 * @return Passt der Wuerfel?
	 */
	private boolean fits(Wuerfel wuerfel, int[] maske) {
		for (int i = 0; i < 6; i++) {
			if (maske[i] != -1 && wuerfel.getWerte()[i] != maske[i]) {
				return false;
			}
		}
		return true;
	}


	/**
	 * Ermittelt die naechste Position, die im Quader befuellt wird
	 *
	 * @param quader         quader des Puzzles
	 * @param letztePosition x,y,z Indizes der letzten befuellten Positionen (3 Eintraege)
	 * @return x, y, z-Indizes der naechsten Position (3 Eintraege)
	 */
	private int[] findPosition(Wuerfel[][][] quader, int[] letztePosition) {
		int puzzleX = holzPuzzle.getQuader().length;
		int puzzleY = holzPuzzle.getQuader()[0].length;
		int puzzleZ = holzPuzzle.getQuader()[0][0].length;
		int x = letztePosition[0];
		int y = letztePosition[1];
		int z = letztePosition[2];
		// {0, 0, 0}
		// {0, 0, 1}
		// {0, 0, 2}
		// {0, 1, 0}
		//''''''
		if (z == puzzleZ - 1) {
			if (y == puzzleY - 1) {
				if (x == puzzleX - 1) {
					// Puzzle ist geloest
					return new int[]{puzzleX, puzzleY, puzzleZ};
				} else {
					return new int[]{x + 1, 0, 0};
				}
			} else {
				return new int[]{x, y + 1, 0};
			}
		} else {
			return new int[]{x, y, z + 1};
		}
	}


	/**
	 * Erstellt ein Schema in das der einzufuegende Wuerfel passen soll.
	 * Beliebige Felder sind mit -1 markiert
	 *
	 * @param quader   Form des Puzzles
	 * @param position Position im Quader, die befuellt werden soll
	 * @return Schema (3 Eintraege)
	 */
	private int[] getMaske(Wuerfel[][][] quader, int[] position) {
		int[] maske = new int[6];
		//Default wert
		Arrays.fill(maske, -1);
		//Seite 3
		if (position[0] == quader.length - 1) {
			maske[2] = 0;
		} else if (quader[position[0] + 1][position[1]][position[2]] != null) {
			if (quader[position[0] + 1][position[1]][position[2]].getWerte()[4] == 1) {
				maske[2] = 2;
			} else if (quader[position[0] + 1][position[1]][position[2]].getWerte()[4] == 2) {
				maske[2] = 1;
			} else if (quader[position[0] + 1][position[1]][position[2]].getWerte()[4] == 3) {
				maske[2] = 4;
			} else if (quader[position[0] + 1][position[1]][position[2]].getWerte()[4] == 4) {
				maske[2] = 3;
			}
		}
		//Seite 5
		if (position[0] == 0) {
			maske[4] = 0;
		} else if (quader[position[0] - 1][position[1]][position[2]] != null) {
			if (quader[position[0] - 1][position[1]][position[2]].getWerte()[2] == 1) {
				maske[4] = 2;
			} else if (quader[position[0] - 1][position[1]][position[2]].getWerte()[2] == 2) {
				maske[4] = 1;
			} else if (quader[position[0] - 1][position[1]][position[2]].getWerte()[2] == 3) {
				maske[4] = 4;
			} else if (quader[position[0] - 1][position[1]][position[2]].getWerte()[2] == 4) {
				maske[4] = 3;
			}
		}
		//Seite 4
		if (position[1] == quader[0].length - 1) {
			maske[3] = 0;
		} else if (quader[position[0]][position[1] + 1][position[2]] != null) {
			if (quader[position[0]][position[1] + 1][position[2]].getWerte()[1] == 1) {
				maske[3] = 2;
			} else if (quader[position[0]][position[1] + 1][position[2]].getWerte()[1] == 2) {
				maske[3] = 1;
			} else if (quader[position[0]][position[1] + 1][position[2]].getWerte()[1] == 3) {
				maske[3] = 4;
			} else if (quader[position[0]][position[1] + 1][position[2]].getWerte()[1] == 4) {
				maske[3] = 3;
			}
		}
		//Seite 2
		if (position[1] == 0) {
			maske[1] = 0;
		} else if (quader[position[0]][position[1] - 1][position[2]] != null) {
			if (quader[position[0]][position[1] - 1][position[2]].getWerte()[3] == 1) {
				maske[1] = 2;
			} else if (quader[position[0]][position[1] - 1][position[2]].getWerte()[3] == 2) {
				maske[1] = 1;
			} else if (quader[position[0]][position[1] - 1][position[2]].getWerte()[3] == 3) {
				maske[1] = 4;
			} else if (quader[position[0]][position[1] - 1][position[2]].getWerte()[3] == 4) {
				maske[1] = 3;
			}
		}
		//Seite 1
		if (position[2] == quader[0][0].length - 1) {
			maske[0] = 0;
		} else if (quader[position[0]][position[1]][position[2] + 1] != null) {
			if (quader[position[0]][position[1]][position[2] + 1].getWerte()[5] == 1) {
				maske[0] = 4;
			} else if (quader[position[0]][position[1]][position[2] + 1].getWerte()[5] == 2) {
				maske[0] = 3;
			} else if (quader[position[0]][position[1]][position[2] + 1].getWerte()[5] == 3) {
				maske[0] = 2;
			} else if (quader[position[0]][position[1]][position[2] + 1].getWerte()[5] == 4) {
				maske[0] = 1;
			}
		}
		//Seite 6
		if (position[2] == 0) {
			maske[5] = 0;
		} else if (quader[position[0]][position[1]][position[2] - 1] != null) {
			if (quader[position[0]][position[1]][position[2] - 1].getWerte()[0] == 1) {
				maske[5] = 4;
			} else if (quader[position[0]][position[1]][position[2] - 1].getWerte()[0] == 2) {
				maske[5] = 3;
			} else if (quader[position[0]][position[1]][position[2] - 1].getWerte()[0] == 3) {
				maske[5] = 2;
			} else if (quader[position[0]][position[1]][position[2] - 1].getWerte()[0] == 4) {
				maske[5] = 1;
			}
		}
		return maske;
	}

	private int getMaskeForChristophOhneVodoo(Wuerfel[][][] quader, int[] position) {
		int result = 0;
		//unterer Nachbar
		if (position[2] != 0 && quader[position[0]][position[1]][position[2] - 1] != null) {
			result += fitsMap.get(50 + quader[position[0]][position[1]][position[2] - 1].getWerte()[0])*100;
		}
		//linker Nachbar
		if (position[1] != 0 && quader[position[0]][position[1]-1][position[2]] != null) {
			result += fitsMap.get(10 + quader[position[0]][position[1] - 1][position[2]].getWerte()[3])*10;
		}
		//hinterer Nachbar
		if (position[0] != 0 && quader[position[0] - 1][position[1]][position[2]] != null) {
			result += fitsMap.get(40 + quader[position[0] - 1][position[1]][position[2]].getWerte()[2]);
		}
		//oben rechts vorne
		if (position[2] == quader[0][0].length - 1) {
			result+= 100000;
		}
		if (position[1] == quader[0].length - 1) {
			result += 10000;
		}
		if (position[0] == quader.length - 1) {
			result += 1000;
		}
		return result;
	}
	/**
	 * Erstellt ein Schema in das der einzufuegende Wuerfel passen soll.
	 * Beliebige Felder sind mit -1 markiert
	 *
	 * @param quader   Form des Puzzles
	 * @param position Position im Quader, die befuellt werden soll
	 * @return Schema (3 Eintraege)
	 */
	private int[] getMaskeOptimized(Wuerfel[][][] quader, int[] position) {
		int[] maske = new int[6];
		//Default wert
		Arrays.fill(maske, -1);
		//maske 0
		if (position[2] == quader[0][0].length - 1) {
			maske[0] = 0;
		} else if (quader[position[0]][position[1]][position[2] + 1] != null) {
			maske[0] = fitsMap.get(quader[position[0]][position[1]][position[2] + 1].getWerte()[5]);
		}
		//maske 1
		if (position[1] == 0) {
			maske[1] = 0;
		} else if (quader[position[0]][position[1] - 1][position[2]] != null) {
			maske[1] = fitsMap.get(10 + quader[position[0]][position[1] - 1][position[2]].getWerte()[3]);
		}
		//maske 2
		if (position[0] == quader.length - 1) {
			maske[2] = 0;
		} else if (quader[position[0] + 1][position[1]][position[2]] != null) {
			maske[2] = fitsMap.get(20 + quader[position[0] + 1][position[1]][position[2]].getWerte()[4]);
		}
		//maske 3
		if (position[1] == quader[0].length - 1) {
			maske[3] = 0;
		} else if (quader[position[0]][position[1] + 1][position[2]] != null) {
			maske[3] = fitsMap.get(30 + quader[position[0]][position[1] + 1][position[2]].getWerte()[1]);
		}
		//maske 4
		if (position[0] == 0) {
			maske[4] = 0;
		} else if (quader[position[0] - 1][position[1]][position[2]] != null) {
			maske[4] = fitsMap.get(40 + quader[position[0] - 1][position[1]][position[2]].getWerte()[2]);
		}
		//maske 5
		if (position[2] == 0) {
			maske[5] = 0;
		} else if (quader[position[0]][position[1]][position[2] - 1] != null) {
			maske[5] = fitsMap.get(50 + quader[position[0]][position[1]][position[2] - 1].getWerte()[0]);
		}
		return maske;
	}

}
