package implementation;

import java.util.*;

public class PuzzleLoeserOptiMopti {
	private HolzPuzzle holzPuzzle;

	/**
	 * Verarbeitet das Holzpuzzle und versucht es durch geschicktes Ausschliessen zu loesen
	 *
	 * @param holzPuzzle ungeloestes Holzpuzzle
	 * @return geloestes Holzpuzzle oder null (bei keiner Loesung)
	 */
	public HolzPuzzle process(HolzPuzzle holzPuzzle) {
		this.holzPuzzle = holzPuzzle;
		List<Wuerfel>[] klassen = klassifiziere();
		if (!Loesbarkeitspruefung.isSolvableRegardingClasses(this.holzPuzzle, klassen)) {
			return null;
		}
		return belelePuzzle(holzPuzzle.getQuader(), new int[]{0, 0, 0}, Builder.buildMap(holzPuzzle.getTeile()));
	}

	/**
	 * Klassifiziere die Wuerfel aus der Liste anhand ihrer Aussenseiten
	 *
	 * @return Array aus Wuerfelklassen
	 */
	private List<Wuerfel>[] klassifiziere() {
		List<Wuerfel>[] klassen = new List[7];
		for (int i = 0; i < 7; i++) {
			klassen[i] = new ArrayList<>();
		}
		for (Wuerfel wuerfel : this.holzPuzzle.getTeile()) {
			klassen[wuerfel.getAussenflaechen()].add(wuerfel);
		}
		return klassen;
	}

	private HolzPuzzle belelePuzzle(Wuerfel[][][] quader, int[] p, Map<Integer, Collection<Builder.WuerfelKlassifikation>> map) {
		if (p[0] == quader.length) {
			return new HolzPuzzle(this.holzPuzzle.getTeile(), quader, this.holzPuzzle.getKommentare());
		}
		int maske = getVodooMaske(quader, p);
		for (Builder.WuerfelKlassifikation option : map.getOrDefault(maske, Collections.emptyList())) {
			if (option.hasEntries()) {
				quader[p[0]][p[1]][p[2]] = option.pop();
				HolzPuzzle puzzle = belelePuzzle(quader, findPosition(quader, p), map);
				if (puzzle != null) {
					return puzzle;
				}
			  quader[p[0]][p[1]][p[2]] = null;//ist ueberfluessig
				option.replenish();
			}
		}
		return null;
	}

	/**
	 * das gegenstueck zu {@link Builder#getKey(Wuerfel)}
	 *
	 * zahl hat folgendes bitschema:
	 * abcdddeeefff
	 *
	 * a: 0 wenn hinten aussenflaeche ist, 1 wenn hinten ein Wuerfel zu erwarten ist
	 * b: wie a fuer rechts
	 * c: wie a fuer links
	 * d: 0 wenn vorne aussenflaeche ist, sonst eine zahl die an das hintere dreieck des vorderen nachbars passt
	 * e: wie d fuer links
	 * f: wie d fuer unoben
	 */
	private int getVodooMaske(Wuerfel[][][] quader, int[] p) {
		// laengen finden
		int vorneHinten = holzPuzzle.getQuader().length;
		int rechtsLinks = holzPuzzle.getQuader()[0].length;
		int obenUnten = holzPuzzle.getQuader()[0][0].length;
		// immer das vorherige Anrainerdreieck finden
		// 0 wenn Aussenflaeche
		int f = p[2] == 0 ? 0 : quader[p[0]][p[1]][p[2] - 1].getWerte()[0];
		int e = p[1] == 0 ? 0 : quader[p[0]][p[1] - 1][p[2]].getWerte()[3];
		int d = p[0] == 0 ? 0 : quader[p[0] - 1][p[1]][p[2]].getWerte()[2];
		// wenn die Flaeche danach aussen ist, 1 sonst ist da noch ein leerer platz, dann 0
		int a = p[0] == vorneHinten -1? 0 : 1;
		int b = p[1] == rechtsLinks -1? 0 : 1;
		int c = p[2] == obenUnten - 1? 0 : 1;
		// alle in einer zahl zusammenfuehren
		return a * 100_000 + b * 10_000 + c * 1000 + d * 100 + e * 10 + f;
	}

	/**
	 * Ermittelt die naechste Position, die im Quader befuellt wird
	 *
	 * @param quader         quader des Puzzles
	 * @param letztePosition x,y,z Indizes der letzten befuellten Positionen (3 Eintraege)
	 * @return x, y, z-Indizes der naechsten Position (3 Eintraege)
	 */
	private int[] findPosition(Wuerfel[][][] quader, int[] letztePosition) {
		int puzzleX = holzPuzzle.getQuader().length;
		int puzzleY = holzPuzzle.getQuader()[0].length;
		int puzzleZ = holzPuzzle.getQuader()[0][0].length;
		int x = letztePosition[0];
		int y = letztePosition[1];
		int z = letztePosition[2];
		// {0, 0, 0}
		// {0, 0, 1}
		// {0, 0, 2}
		// {0, 1, 0}
		//''''''
		if (z == puzzleZ - 1) {
			if (y == puzzleY - 1) {
				if (x == puzzleX - 1) {
					// Puzzle ist geloest
					return new int[]{puzzleX, puzzleY, puzzleZ};
				} else {
					return new int[]{x + 1, 0, 0};
				}
			} else {
				return new int[]{x, y + 1, 0};
			}
		} else {
			return new int[]{x, y, z + 1};
		}
	}
}
